package dche;
import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.io.File;
import java.io.FileReader;
import java.io.InputStream;
import java.net.URL;
import dche.component.DcheActionManager;
import dche.component.DcheViewCanvas;
import dche.io.DcheModelReader;
import dche.io.DcheResultWriter;
import figure.matrix.MatrixSolveMethod;

//--------------------------------------------------------------------//
/**
 * DCHE V~[VAvbg CNX
 * @author aG
 * @version 0.2
 */
public class DcheApplet extends Applet{

/** G[bZ[W */
  protected String message;

//--------------------------------------------------------------------//
/**
 * C\bh<br>
 * w肵t@Cf[^ǂݍ݌vZߓ_xo͂
 * @param args R}hIvVi[z
 */
  public static void main(String args[]){
    DcheDataModel model=new DcheDataModel();
    try{
      DcheModelReader reader=new DcheModelReader(model);
      reader.read(new FileReader(new File(args[0])));
      reader=null;
      model.calculate();
      DcheResultWriter writer=new DcheResultWriter(model);
      writer.write(args[1]);
      writer=null;
    }
    catch(Exception ex){
      ex.printStackTrace();
    }
  }

//--------------------------------------------------------------------//
/**
 * Avbg
 */
  public void init(){
    setLayout(new BorderLayout());
    Color bgColor=getTagColor("backgroundColor");
    Color fgColor=getTagColor("foregroundColor");
    DcheDataModel model=getDataModel();
    model.setMethod(getMatrixSolveMethod());
    DcheViewCanvas canvas=new DcheViewCanvas
      (model,message,getTagFont(),bgColor,fgColor);
    DcheActionManager manager=new DcheActionManager(canvas);
    add(canvas,BorderLayout.CENTER);
    add(manager.getControlPanel(),BorderLayout.SOUTH);
  }

//--------------------------------------------------------------------//
/**
 * \[Xj
 */
  public void destroy(){
    removeAll();
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽ FEM f[^Ԃ
 * @return HTML^OŎw肳ꂽ FEM f[^
 */
  public DcheDataModel getDataModel(){
    String dataName=getParameter("data");
    if(dataName==null) dataName="sample.dat";
    message=null;
    try{
      Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
    }
    catch(SecurityException ex){
      message=String.valueOf(ex);
      ex.printStackTrace();
      return null;
    }
    DcheDataModel model=new DcheDataModel();
    try{
      DcheModelReader reader=new DcheModelReader(model);
      reader.read(new URL(getDocumentBase(),dataName));
      return model;
    }
    catch(Exception ex){
      message=String.valueOf(ex);
      ex.printStackTrace();
      return null;
    }
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽFԂ
 * @param name p[^
 * @return HTML^OŎw肳ꂽF
 */
  protected Color getTagColor(String name){
    String value=getParameter(name);
    if(value==null) return null;
    try{
      return new Color(Integer.parseInt(value,16));
    }
    catch(NumberFormatException ex){
      return null;
    }
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽtHgԂ
 * @return HTML^OŎw肳ꂽtHg
 */
  protected Font getTagFont(){
    String fontName=getParameter("fontName");
    if(fontName==null) fontName="Serif";

    String s=getParameter("fontStyle");
    int fontStyle=Font.PLAIN;
    if(s!=null){
      if(s.equals("bold")){
      	fontStyle=Font.BOLD;
      }
      else if(s.equals("italic")){
      	fontStyle=Font.ITALIC;
      }
    }

    s=getParameter("fontSize");
    int fontSize=14;
    try{
      fontSize=Integer.parseInt(s);
    }
    catch(NumberFormatException ex){
    }

    return new Font(fontName,fontStyle,fontSize);
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽs@Ԃ
 * @return HTML^OŎw肳ꂽs@
 */
  protected MatrixSolveMethod getMatrixSolveMethod(){
    MatrixSolveMethod method=MatrixSolveMethod.ILUCG;
    String value=getParameter("method");
    if(value==null){
      return method;
    }
    else if(value.equals(MatrixSolveMethod.GAUSS.getName())){
      method=MatrixSolveMethod.GAUSS;
    }
    else if(value.equals(MatrixSolveMethod.GAUSS_JORDAN.getName())){
      method=MatrixSolveMethod.GAUSS_JORDAN;
    }
    else if(value.equals(MatrixSolveMethod.LU.getName())){
      method=MatrixSolveMethod.LU;
    }
    else if(value.equals(MatrixSolveMethod.SOR.getName())){
      method=MatrixSolveMethod.SOR;
    }
    else if(value.equals(MatrixSolveMethod.CG.getName())){
      method=MatrixSolveMethod.CG;
    }
    return method;
  }

//--------------------------------------------------------------------//
/**
 * HTML^OŎw肳ꂽlԂ
 * @param name p[^
 * @return HTML^OŎw肳ꂽl
 */
  protected double getTagNumber(String name){
    String value=getParameter(name);
    if(value==null) return Double.NaN;
    try{
      return Double.parseDouble(value);
    }
    catch(NumberFormatException ex){
      return Double.NaN;
    }
  }

//--------------------------------------------------------------------//
/**
 * AvbgԂ
 * @return Avbg܂ޕ
 */
  public String getAppletInfo(){
    return "Applet of DCHE simulation\nAuthor K.Higuchi";
  }

//--------------------------------------------------------------------//
/**
 * ̃Avbg߂p[^Ԃ
 * @return Avbg߂p[^i[z
 */
  public String[][] getParameterInfo(){
    return new String[][]
      {{"data","String","DCHE model file name."},
       {"fontName","String","Font name. Default is Serif."},
       {"fontStyle","int","Font style. Default is Font.PLAIN."},
       {"fontSize","int","Font size. Default is 14."},
       {"backgroundColor","hexadecimal RGB number",
      	"The color of background."},
       {"foregroundColor","hexadecimal RGB number",
      	"The color of foreground."},
       {"method","String",
      	"The method for solving systems of linear equations."}};
  }
}
