package dche;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import fem.Element2D;
import fem.ElementBorder;
import fem.FEMDataModel;
import fem.LabelException;
import fem.LabelMap;
import fem.Material;
import fem.MeshModel;
import fem.Node;
import fem.Nodes;
import fem.bc.BoundaryCondition;
import fem.bc.HeatFluxBound;
import fem.bc.HeatTransferBound;
import fem.bc.Temperature;
import fem.result.Result;
import figure.Matrix;
import figure.Vector;
import figure.matrix.MatrixSolveException;
import figure.matrix.MatrixSolveMethod;
import figure.matrix.MatrixSolver;
import figure.matrix.SparseMatrixSolver;

//--------------------------------------------------------------------//
/**
 * DCHE f[^f
 */
public class DcheDataModel extends FEMDataModel{

/** vZ񐔂̏ */
  public static final int ITER_MAX=100;

/** 臒l */
  public static final double TOLERANCE=1.0e-8;

/** ̕i[郊Xg */
  protected ArrayList<Fluid> fluids;

/** DCHE */
  protected Dche dche;

/** j^[ߓ_i[z */
  protected int monitor[];

/**
 * DCHE f[^f\z
 */
  public DcheDataModel(){
    super();
    fluids=new ArrayList<Fluid>();
    dche=null;
    monitor=new int[0];
  }

//--------------------------------------------------------------------//
/**
 * ̐̕Ԃ
 * @return ̐̕
 */
  public int fluidCount(){
    return fluids.size();
  }

//--------------------------------------------------------------------//
/**
 * ̕Ԃ
 * @param index ̃̕CfbNX
 * @return w肳ꂽʒu̗̕
 */
  public Fluid getFluid(int index){
    return fluids.get(index);
  }

//--------------------------------------------------------------------//
/**
 * ̕ǉ
 * @param fluid ̕
 */
  public void addFluid(Fluid fluid){
    fluids.add(fluid);
  }

//--------------------------------------------------------------------//
/**
 * DCHE Ԃ
 * @return DCHE
 */
  public Dche getDche(){
    return dche;
  }

//--------------------------------------------------------------------//
/**
 * DCHE ݒ肷
 * @param dche DCHE
 */
  public void setDche(Dche dche){
    this.dche=dche;
  }

//--------------------------------------------------------------------//
/**
 * j^[ߓ_ݒ肷
 * @param monitor j^[ߓ_ԍi[z
 */
  public void setMonitor(int monitor[]){
    this.monitor=monitor;
  }

//--------------------------------------------------------------------//
/**
 * j^[ߓ_Ԃ
 * @return j^[ߓ_ԍi[z
 */
  public int[] getMonitor(){
    return monitor;
  }

//--------------------------------------------------------------------//
/**
 * f[^
 */
  public void clear(){
    super.clear();
    fluids.clear();
    monitor=new int[0];
  }

//--------------------------------------------------------------------//
/**
 * vZʂ
 */
  public void clearResults(){
    super.clearResults();
    result.fillTemperature(mesh.getNodeCount(),20.0);
  }

//--------------------------------------------------------------------//
/**
 * f
 */
  public void init() throws LabelException{
    materials.trimToSize();
    fluids.trimToSize();
    Collections.sort(materials,MeshModel.LABEL_COMPARATOR);
    mesh.init();
    bc.init();

    reNumbering();
    resetMaterialLabel();
    resetFluidLabel();
    mesh.getFreeBorder();
    modelExist=true;
    result.fillTemperature(mesh.getNodeCount(),20.0);
  }

//--------------------------------------------------------------------//
/**
 * ߓ_|C^ݒ肷
 * @throws LabelException f[^xvȂƂ
 */
  public void reNumbering() throws LabelException{
    LabelMap labelMap=new LabelMap();
    int nodeCount=mesh.getNodeCount();
    for(int i=0;i<nodeCount;i++){
      labelMap.put(mesh.getNode(i),i);
    }

    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      labelMap.reset((Nodes)mesh.getElement(i));
    }

    int tempCount=bc.temperatureCount();
    for(int i=0;i<tempCount;i++){
      labelMap.reset(bc.getTemperature(i));
    }
    if(dche!=null){
      int count=dche.nodeCount();
      for(int i=0;i<count;i++){
      	Integer newLabel=labelMap.get(dche.node(i));
      	if(newLabel!=null){
      	  dche.resetNodeLabel(i,newLabel.intValue());
      	}
      	else{
      	  throw new LabelException("Bad node label");
      	}
      	newLabel=labelMap.get(dche.node2(i));
      	if(newLabel!=null){
      	  dche.resetNodeLabel2(i,newLabel.intValue());
      	}
      	else{
      	  throw new LabelException("Bad node label");
      	}
      	newLabel=labelMap.get(dche.node3(i));
      	if(newLabel!=null){
      	  dche.resetNodeLabel3(i,newLabel.intValue());
      	}
      	else{
      	  throw new LabelException("Bad node label");
      	}
      }
    }
    if(monitor.length>0){
      for(int i=0;i<monitor.length;i++){
      	Integer newLabel=labelMap.get(monitor[i]);
      	if(newLabel!=null){
      	  monitor[i]=newLabel.intValue();
     	 }
      	else{
      	  throw new LabelException("Bad node label");
      	}
      }
    }

    labelMap.clear();
    for(int i=0;i<elementCount;i++){
      labelMap.put(mesh.getElement(i),i);
    }

    int htcCount=bc.htcCount();
    for(int i=0;i<htcCount;i++){
      labelMap.reset(bc.getHTBound(i));
    }
    int fluxCount=bc.fluxCount();
    for(int i=0;i<fluxCount;i++){
      labelMap.reset(bc.getFluxBound(i));
    }
  }

//--------------------------------------------------------------------//
/**
 * ̃|C^ݒ肷
 * @throws LabelException f[^xvȂƂ
 */
  public void resetFluidLabel() throws LabelException{
    if(dche==null) return;
    HashMap<Integer,Integer> fluidMap=new HashMap<Integer,Integer>();
    int fluidCount=fluids.size();
    for(int i=0;i<fluidCount;i++){
      fluidMap.put(getFluid(i).label(),i);
    }
    if(dche!=null){
      Integer newFluid=fluidMap.get(dche.getFluid());
      if(newFluid!=null){
      	dche.setFluid(newFluid.intValue());
      }
    }
    else{
      throw new LabelException("Bad fluid label");
    }
  }

//--------------------------------------------------------------------//
/**
 * M`BvZ
 */
  public void calculate(){
    double t0[]=result.getTemperatures();
    double dt=1.0;
    int iter=0;
    try{
      do{
      	createHeatMatrix();
      	Vector temp=matrix.solve(vector,method);
      	result.clear();
      	result.setTemperature(bc,temp,bc.getTemperature());
      	double t1[]=result.getTemperatures();
      	dt=0.0;
      	for(int i=0;i<t0.length;i++){
      	  dt=Math.max(dt,Math.abs(t1[i]-t0[i]));
      	}
      	iter++;
      	t0=t1;
      	System.out.println(iter+":"+dt);
      }while((iter<ITER_MAX) && (dt>=TOLERANCE));
      if(monitor.length>0){
      	for(int i=0;i<monitor.length;i++){
      	  int label=mesh.getNode(monitor[i]).label();
      	  System.out.println
      	    (label+" : "+result.getTemperature(monitor[i]));
      	}
      }
    }
    catch(MatrixSolveException ex){
      ex.printStackTrace();
    }
  }

//--------------------------------------------------------------------//
/**
 * MvZ̃}gbNXExNgvZ
 */
  protected void createHeatMatrix(){
    int nodeCount=mesh.getNodeCount();
    int elementCount=mesh.getElementCount();

    int dof=bc.bcList(mesh.getNodeCount());
    int length=nodeCount-dof;
    int ii=0;
    int reducedList[]=new int[length];
    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]<0){
      	reducedList[ii]=i;
      	ii++;
      }
    }

    SparseMatrixSolver matrix1=new SparseMatrixSolver(nodeCount);
    Vector vector1=bc.loadVector(nodeCount);
    for(int i=0;i<elementCount;i++){
      Element2D element=(Element2D)mesh.getElement(i);
      int ncount=element.nodeCount();
      int nodesOnElement[]=element.nodes();
      Material material=materials.get(element.getMaterial());
      Matrix ls=element.gradMatrix(mesh.nodes(element),material);
      for(int i1=0;i1<ncount;i1++){
      	for(int i2=0;i2<ncount;i2++){
      	  matrix1.add(nodesOnElement[i1],nodesOnElement[i2],
      	      	      ls.get(i1,i2));
      	}
      }
    }

    int boundCount=bc.htcCount();
    for(int i=0;i<boundCount;i++){
      HeatTransferBound htc=bc.getHTBound(i);
      ElementBorder border=htc.getBorder(mesh.getElement(htc.label()));
      int ncount=border.nodeCount();
      int nodesOnElement[]=border.nodes();
      Node nodes[]=mesh.nodes(border);
      Matrix hm=border.shapeFunctionMatrix(nodes,htc.htc);
      Vector hv=border.shapeFunctionVector(nodes,htc.outTemp*htc.htc);
      for(int i1=0;i1<ncount;i1++){
      	for(int i2=0;i2<ncount;i2++){
      	  matrix1.add(nodesOnElement[i1],nodesOnElement[i2],
      	      	      hm.get(i1,i2));
      	}
      	vector1.add(nodesOnElement[i1],hv.get(i1));
      }
    }
    boundCount=bc.fluxCount();
    for(int i=0;i<boundCount;i++){
      HeatFluxBound flux=bc.getFluxBound(i);
      ElementBorder border=flux.getBorder(mesh.getElement(flux.label()));
      int ncount=border.nodeCount();
      int nodesOnElement[]=border.nodes();
      Node nodes[]=mesh.nodes(border);
      Vector hv=border.shapeFunctionVector(nodes,flux.flux);
      for(int i1=0;i1<ncount;i1++){
      	vector1.add(nodesOnElement[i1],hv.get(i1));
      }
    }
    if(dche!=null){
      int ncount=dche.nodeCount();
      int nodes1[]=dche.nodes();
      Node p[]=new Node[ncount];
      for(int i=0;i<ncount;i++){
      	p[i]=mesh.getNode(nodes1[i]);
      }
      Fluid fl=getFluid(dche.getFluid());
      Matrix hm=dche.calc(result,fl,p);
      for(int i1=0;i1<ncount;i1++){
      	int node3=dche.node3(i1);
      	matrix1.add(nodes1[i1],node3,hm.get(i1,2*ncount+i1-1));
      	matrix1.add(node3,nodes1[i1],hm.get(2*ncount+i1-1,i1));
      	matrix1.add(node3,node3,hm.get(2*ncount+i1-1,2*ncount+i1-1));
      	for(int i2=0;i2<ncount;i2++){
      	  matrix1.add(nodes1[i1],nodes1[i2],hm.get(i1,i2));
      	}
      }
      for(int i1=0;i1<ncount-1;i1++){
      	int node2=dche.node2(i1);
      	matrix1.add(nodes1[i1],node2,hm.get(i1,ncount+i1));
      	matrix1.add(node2,nodes1[i1],hm.get(ncount+i1,i1));
      	matrix1.add(node2,node2,hm.get(ncount+i1,ncount+i1));
      }
      for(int i1=0;i1<ncount;i1++){
      	int node2=dche.node2(i1);
      	int j1=ncount+i1;
      	if(i1==ncount-1) j1=i1;
      	for(int i2=0;i2<ncount;i2++){
      	  if(i1!=i2){
      	    int node22=dche.node2(i2);
      	    int j2=ncount+i2;
      	    if(i2==ncount-1) j2=i2;
      	    matrix1.add(node2,node22,hm.get(j1,j2));
      	  }
      	}
      }
    }
    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]>=0){
      	Temperature temp=bc.getTemperature(bc.bcList[i]);
      	for(int j=0;j<vector1.size;j++){
      	  vector1.add(j,-temp.temperature()*matrix1.get(j,i));
      	}
      }
    }

    matrix=matrix1.extruct(length,reducedList,reducedList);
    vector=new Vector(length);
    for(int i=0;i<length;i++){
      vector.set(i,vector1.get(reducedList[i]));
    }
  }
}
