package dche;
import fem.LabeledData;
import math.CubicFunction;
import math.Function;

//--------------------------------------------------------------------//
/**
 * ̕NX
 */
public class Fluid implements LabeledData{

/** ̔ԍ */
  public int label;

/** xf[^̐ */
  public int count;

/** xi[z */
  protected double t[];

/** vfi[z */
  protected FluidElement elem[];

/** l֐i[z */
  protected Function functions[][];

/**
 * ̕\z
 * @param label ̔ԍ
 * @param count xf[^̐
 * @param t xi[z
 * @param elem vfi[z
 */
  public Fluid(int label,int count,double t[],FluidElement elem[]){
    this.label=label;
    this.count=count;
    this.t=t;
    this.elem=elem;
    setCoefficient();
  }

//--------------------------------------------------------------------//
/**
 * ̔ԍԂ
 * @return ̔ԍ
 */
  public int label(){
    return label;
  }

//--------------------------------------------------------------------//
/**
 * ̔ԍݒ肷
 * @param label ̔ԍ
 */
  public void setLabel(int label){
    this.label=label;
  }

//--------------------------------------------------------------------//
/**
 * Wݒ肷
 */
  protected void setCoefficient(){
    if(count<=1) return;
    functions=new Function[5][count-1];
    double y[][]=new double[5][count];
    double h[]=new double[count-1];
    double hinv[]=new double[count-1];
    double balpha[]=new double[count-1];
    double gamma[]=new double[count-1];
    double xm[]=new double[count-1];

    for(int i=0;i<count;i++){
      y[0][i]=elem[i].density;
      y[1][i]=elem[i].specificHeat;
      y[2][i]=elem[i].heatConductivity;
      y[3][i]=elem[i].viscocity;
      y[4][i]=elem[i].prandtl;
    }
    for(int i=0;i<count-1;i++){
      h[i]=t[i+1]-t[i];
      hinv[i]=1.0/h[i];
    }
    gamma[0]=0.0;
    xm[count-2]=0.0;
    double bb=h[1]*h[0]+1.0;
    balpha[1]=0.5/bb;
    for(int i=2;i<count-1;i++){
      gamma[i-1]=(bb-1.0)*balpha[i-1];
      bb=h[i]*hinv[i-1]+1.0;
      balpha[i]=1.0/(bb+bb-gamma[i-1]);
    }
    for(int j=0;j<5;j++){
      double e1=(y[j][1]-y[j][0])*hinv[0];
      double e2=(y[j][2]-y[j][1])*hinv[1];
      xm[0]=(e2-e1)*hinv[0]*balpha[1];
      for(int i=2;i<count-1;i++){
      	e1=e2;
      	e2=(y[j][i+1]-y[j][i])*hinv[i];
      	xm[i-1]=((e2-e1)*hinv[i-1]-xm[i-2])*balpha[i];
      }
      for(int i=count-4;i>=0;i--) xm[i]=xm[i]-gamma[i]*xm[i+1];
      double a30=xm[0];
      double a20=xm[0]*t[0];
      double a10=a20*t[0];
      double a3=a30*hinv[0];
      double a2=-3.0*a20*hinv[0];
      double a1=(3.0*a10+y[j][1]-y[j][0])*hinv[0]-a30*h[0];
      double a0=(y[j][0]*t[1]-(a10+y[j][1])*t[0])*hinv[0]+a20*h[0];
      functions[j][0]=new CubicFunction(a3,a2,a1,a0);
      for(int i=1;i<count-1;i++){
      	a30=xm[i]-xm[i-1];
      	double a21=xm[i-1]*t[i+1];
      	double a22=xm[i]*t[i];
      	double a11=a21*t[i+1];
      	double a12=a22*t[i];
      	double a00=(a11+y[j][i])*t[i+1]-(a12+y[j][i+1])*t[i];
      	a3=a30*hinv[i];
      	a2=3.0*(a21-a22)*hinv[i];
      	a1=(3.0*(a12-a11)+y[j][i+1]-y[j][i])*hinv[i]-a30*h[i];
      	a0=a00*hinv[i]-(a21-a22)*h[i];
      	functions[j][i]=new CubicFunction(a3,a2,a1,a0);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * ԂlԂ
 * @param tp x
 * @return Ԃvf
 */
  public FluidElement get(double tp){
    if((count<2) || (tp<=t[0])){
      return elem[0];
    }
    else if(tp>=t[count-1]){
      return elem[count-1];
    }
    else{
      int low=0;
      int high=count-1;
      do{
      	int mid=(low+high)/2;
      	if(t[mid]<=tp) low=mid+1;
      	else           high=mid;
      }while(low<high);
      int i=(low+high)/2-1;
      return new FluidElement(functions[0][i].getValue(tp),
      	      	      	      functions[1][i].getValue(tp),
      	      	      	      functions[2][i].getValue(tp),
      	      	      	      functions[3][i].getValue(tp),
      	      	      	      functions[4][i].getValue(tp));
    }
  }
}
