package dche.component;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import data.NumberString;
import dche.Dche;
import dche.DcheDataModel;
import fem.Element2D;
import fem.ElementBorder;
import fem.MeshModel;
import fem.Node;
import fem.bc.BoundaryCondition;
import fem.bc.HeatTransferBound;
import fem.bc.Temperature;
import fem.result.ContourData;
import fem.result.Result;
import figure.Transform;
import gui.BasicViewCanvas;
import gui.ColorBar;

//--------------------------------------------------------------------//
/**
 * DCHE f\LoX
 */
public class DcheViewCanvas extends BasicViewCanvas{

/** ̍ŏl */
  public static final double MIN_VALUE=1.0E-20;

/** DCHE f[^f */
  protected DcheDataModel model;

/** f\̈TCY */
  protected int viewSize;

/** ʒ_ */
  protected Point centerPoint;

/** G[bZ[W */
  protected String message;

/** tHg */
  protected Font font;

/** wiF */
  protected Color bgColor;

/** OiF */
  protected Color fgColor;

/** vf\F */
  protected Color elemColor;

/** ߓ_x\F */
  protected Color tempColor;

/** M`BE\F */
  protected Color htcColor;

/** fsړW */
  protected double transCoef;

/** f]xW */
  protected double rotateCoef;

/** fgEkW */
  protected double zoomCoef;

/** ϊs */
  protected Transform transform;

/** J[o[ */
  protected ColorBar colorbar;

/** J[o[\̈\ <code>Rectangle</code> */
  protected Rectangle colorbarBound;

/** R^[}f[^ */
  protected ContourData contour;

/**
 * DCHE f\LoX\z
 * @param model FEM f[^f
 */
  public DcheViewCanvas(DcheDataModel model,String message,Font font,
      	      	      	Color bgColor,Color fgColor){
    this.model=model;
    this.message=message;
    this.font=font;
    setColors(bgColor,fgColor);
    elemColor=Color.GRAY;
    tempColor=Color.BLUE;
    htcColor=Color.MAGENTA;
    transform=new Transform();
    colorbar=null;
    contour=null;
  }

//--------------------------------------------------------------------//
/**
 * \Fݒ肷
 * @param bgColor wiF
 * @param fgColor OiF
 */
  protected void setColors(Color bgColor,Color fgColor){
    if(bgColor==null){
      this.bgColor=Color.WHITE;
    }
    else{
      this.bgColor=bgColor;
    }
    if(fgColor==null){
      this.fgColor=Color.BLACK;
    }
    else{
      this.fgColor=fgColor;
    }
    setBackground(bgColor);
  }

//--------------------------------------------------------------------//
/**
 * DCHE f[^fԂ
 * @return DCHE f[^f
 */
  public DcheDataModel getModel(){
    return model;
  }

//--------------------------------------------------------------------//
/**
 * f𕽍sړ
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void translate(int dx,int dy){
    transform.translate(transCoef*(double)dx,-transCoef*(double)dy,0.0);
  }

//--------------------------------------------------------------------//
/**
 * f]
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void rotate(int dx,int dy){
    transform.rotate(rotateCoef*(double)dy,rotateCoef*(double)dx,0.0);
  }

//--------------------------------------------------------------------//
/**
 * fgEk
 * @param dx ʏ <i>x</i> ړ
 * @param dy ʏ <i>y</i> ړ
 */
  public void zoom(int dx,int dy){
    transform.zoom(Math.pow(2.0,zoomCoef*(double)(dx+dy)));
  }

//--------------------------------------------------------------------//
/**
 * R^[\ݒ肷
 */
  public void showResult(){
    if(model.getResult().calculated()){
      contour=new ContourData();
      contour.setContour(model.getResult(),Result.TEMPERATURE);
      colorbar=new ColorBar.Index(contour.colorbar(),20);
    }
    else{
      contour=null;
      colorbar=null;
    }
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * vZʕ\
 */
  public void clear(){
    contour=null;
    colorbar=null;
    repaint();
  }

//--------------------------------------------------------------------//
/**
 * J[o[ݒ肷
 * @param colorbar J[o[
 */
  public void setColorBar(ColorBar colorbar){
    this.colorbar=colorbar;
  }

//--------------------------------------------------------------------//
/**
 * _ʓɂ邩ǂʂ
 * @param x <i>x</i> W
 * @param y <i>y</i> W
 * @return _ʓɂꍇ true
 */
  protected boolean innerFrame(int x,int y){
    return ((0<=x) && (x<panelSize.width)
      	 && (0<=y) && (y<panelSize.height));
  }

//--------------------------------------------------------------------//
/**
 * ĕ`悷
 * @param g `悷OtBbNXReLXg
 */
  public void update(Graphics g){
    if(model!=null){
      paintModel(bg);
    }
    else{
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * `悷
 * @param g `悷OtBbNXReLXg
 */
  public void paint(Graphics g){
    setPaint();
    if(model!=null){
      paintModel(bg);
    }
    else{
      drawErrorMessage(bg,message);
    }
    g.drawImage(bgImage,0,0,this);
  }

//--------------------------------------------------------------------//
/**
 * ʐݒ
 */
  protected void setPaint(){
    super.setPaint();
    viewSize=Math.min(panelSize.width,panelSize.height);
    centerPoint=new Point(viewSize/2,viewSize/2);
    transCoef=1.0;
    rotateCoef=Math.PI/(double)viewSize;
    zoomCoef=1.0/(double)viewSize;
    bg.setFont(font);
    if(panelSize.width>panelSize.height){
      colorbarBound=new Rectangle
      	(panelSize.height*11/10,panelSize.height/10,
      	 panelSize.width/20,panelSize.height*4/5);
    }
    else{
      colorbarBound=new Rectangle
      	(panelSize.width/10,panelSize.width*11/10,
      	 panelSize.width*4/5,panelSize.height/20);
    }
  }

//--------------------------------------------------------------------//
/**
 * `e
 * @param g `悷OtBbNXReLXg
 */
  protected void paintModel(Graphics2D g){
    g.setColor(bgColor);
    g.fillRect(0,0,panelSize.width,panelSize.height);
    if(!model.modelExist()) return;

    Rectangle2D bounds=model.getMesh().getBounds();
    double cx=bounds.getCenterX();
    double cy=bounds.getCenterY();
    double boundSize=Math.max(bounds.getWidth(),bounds.getHeight());
    double vCoef=0.8*(double)viewSize/Math.max(boundSize,MIN_VALUE);

    int nodeCount=model.getMesh().getNodeCount();
    ViewPoint2D vp1[]=new ViewPoint2D[nodeCount];
    for(int i=0;i<nodeCount;i++){
      vp1[i]=getViewPoint(i,vCoef,cx,cy);
    }
    g.setColor(fgColor);
    if(model.getResult().calculated()){
      String tempMax=NumberString.toString(model.getResult().tempMax());
      g.drawString("Temperature: "+tempMax,10,20);
    }
    else{
      int elementCount=model.getMesh().getElementCount();
      g.drawString("Node:    "+nodeCount,10,20);
      g.drawString("Element: "+elementCount,10,40);
    }

    if(contour!=null){
      drawElement(g,vp1,contour,colorbar);
      if(model.getDche()!=null){
      	drawDche(g,vp1,contour,colorbar);
      }
      if(panelSize.width>panelSize.height){
      	colorbar.drawVerticalBar(g,colorbarBound,fgColor);
      }
      else{
      	colorbar.drawHorizontalBar(g,colorbarBound,fgColor);
      }
    }

    drawElement(g,vp1);
    g.setColor(tempColor);
    drawTemperature(g,vp1);
    g.setColor(htcColor);
    drawHTBound(g,vp1);
  }

//--------------------------------------------------------------------//
/**
 * vf`悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 * @param contour R^[}f[^
 * @param colorbar J[o[
 */
  protected void drawElement(Graphics2D g,ViewPoint2D vp[],
      	      	      	     ContourData contour,ColorBar colorbar){
    MeshModel mesh=model.getMesh();
    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      Element2D elem=mesh.getElement(i);
      int count=elem.nodeCount();
      int gx[]=new int[count];
      int gy[]=new int[count];
      double value[]=new double[count];
      for(int j=0;j<count;j++){
      	int node=elem.node(j);
      	gx[j]=vp[node].getX(centerPoint);
      	gy[j]=vp[node].getY(centerPoint);
      	value[j]=contour.getContourData(node);
      }
      fillPolygon(g,count,gx,gy,value,colorbar);
    }
  }

//--------------------------------------------------------------------//
/**
 * DCHE `悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 * @param contour R^[}f[^
 * @param colorbar J[o[
 */
  protected void drawDche(Graphics2D g,ViewPoint2D vp[],
      	      	      	  ContourData contour,ColorBar colorbar){
    Dche dche=model.getDche();
    int count=dche.nodeCount();
    int gx[]=new int[count];
    int gy[]=new int[count];
    double value[]=new double[count];
    for(int i=0;i<count;i++){
      int node=dche.node(i);
      gx[i]=vp[node].getX(centerPoint);
      gy[i]=vp[node].getY(centerPoint);
      value[i]=contour.getContourData(node);
      if(i>0){
      	g.setColor(colorbar.color(0.5*(value[i-1]+value[i])));
      	g.drawLine(gx[i-1],gy[i-1],gx[i],gy[i]);
      	g.drawLine(gx[i-1]-1,gy[i-1],gx[i]-1,gy[i]);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * vf`悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 */
  protected void drawElement(Graphics2D g,ViewPoint2D vp[]){
    g.setColor(fgColor);
    MeshModel mesh=model.getMesh();
    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      Element2D elem=mesh.getElement(i);
      int count=elem.nodeCount();
      int gx[]=new int[count];
      int gy[]=new int[count];
      for(int j=0;j<count;j++){
      	int node=elem.node(j);
      	gx[j]=vp[node].getX(centerPoint);
      	gy[j]=vp[node].getY(centerPoint);
      }
      g.draw(new Polygon(gx,gy,count));
    }
  }

//--------------------------------------------------------------------//
/**
 * ߓ_x`悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 */
  protected void drawTemperature(Graphics2D g,ViewPoint2D vp[]){
    BoundaryCondition bc=model.getBoundaryCondition();
    int temperatureCount=bc.temperatureCount();
    for(int i=0;i<temperatureCount;i++){
      Temperature temperature=bc.getTemperature(i);
      Point g1=vp[temperature.label()].getPoint(centerPoint);
      if(innerFrame(g1.x,g1.y)){
      	g.fillOval(g1.x-2,g1.y-2,4,4);
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * M`BE`悷
 * @param g `悷OtBbNXReLXg
 * @param vp ߓ_̉ʏ̓_i[z
 */
  protected void drawHTBound(Graphics2D g,ViewPoint2D vp[]){
    BoundaryCondition bc=model.getBoundaryCondition();
    double vsize=0.02*(double)viewSize;
    MeshModel mesh=model.getMesh();
    int htcCount=bc.htcCount();
    for(int i=0;i<htcCount;i++){
      HeatTransferBound htc=bc.getHTBound(i);
      ElementBorder border=htc.getBorder(mesh.getElement(htc.label()));
      Point g1=vp[border.node(0)].getPoint(centerPoint);
      Point g2=vp[border.node(1)].getPoint(centerPoint);
      g.drawLine(g1.x,g1.y,g2.x,g2.y);
    }
  }

//--------------------------------------------------------------------//
/**
 * xNg`悷
 * @param g `悷OtBbNXReLXg
 * @param p1 xNg̎n_
 * @param p2 xNg̏I_
 * @param vsize [\TCY
 */
  protected void drawVector(Graphics2D g,Point p1,Point p2,
      	      	      	    double vsize){
    g.drawLine(p1.x,p1.y,p2.x,p2.y);
    double dx=(double)(p1.x-p2.x);
    double dy=(double)(p1.y-p2.y);
    double dd=Math.sqrt(dx*dx+dy*dy);
    if(dd>0.0) vsize/=dd;
    double cc=0.5*vsize*Math.sqrt(3.0);
    double cs=0.5*vsize;
    int x1=p2.x+(int)(cc*dx-cs*dy);
    int y1=p2.y+(int)(cs*dx+cc*dy);
    int x2=p2.x+(int)(cc*dx+cs*dy);
    int y2=p2.y+(int)(-cs*dx+cc*dy);
    g.drawLine(x1,y1,p2.x,p2.y);
    g.drawLine(x2,y2,p2.x,p2.y);
  }

//--------------------------------------------------------------------//
/**
 * R^[p``悷
 * @param g `悷OtBbNXReLXg
 * @param n _̐
 * @param gx _ <i>x</i> Wi[z
 * @param gy _ <i>y</i> Wi[z
 * @param value _̒li[z
 * @param colorbar J[o[
 */
  protected void fillPolygon(Graphics2D g,int n,int gx[],int gy[],
      	      	      	     double value[],ColorBar colorbar){
    Polygon pl=new Polygon(gx,gy,n);
    Rectangle rect=pl.getBounds();
    int xMin=Math.max(0,rect.x);
    int xMax=Math.min(panelSize.width,rect.x+rect.width);
    int yMin=Math.max(0,rect.y);
    int yMax=Math.min(panelSize.height,rect.y+rect.height);
    if((xMin>=xMax) || (yMin>=yMax) || (xMax<0) || (yMax<0)) return;
    byte drawtype[][]=new byte[xMax-xMin+1][yMax-yMin+1];
    for(int i=xMin;i<=xMax;i++){
      for(int j=yMin;j<=yMax;j++){
      	if(pl.contains(i,j)) drawtype[i-xMin][j-yMin]=1;
      	else                 drawtype[i-xMin][j-yMin]=0;
      }
    }
    switch(n){
      case 3:
      	for(int i=xMin;i<xMax;i++){
      	  for(int j=yMin;j<yMax;j++){
      	    if(drawtype[i-xMin][j-yMin]==1){
      	      fillPolygon(g,i,j,value,
      	      	      	  PolygonCoordinates.triangle(i,j,gx,gy));
      	    }
      	  }
      	}
      	break;
      case 4:
      	for(int i=xMin;i<xMax;i++){
      	  for(int j=yMin;j<yMax;j++){
      	    if(drawtype[i-xMin][j-yMin]==1){
      	      fillPolygon(g,i,j,value,
      	      	      	  PolygonCoordinates.quadangle(i,j,gx,gy));
      	    }
      	  }
      	}
      	break;
    }
  }

//--------------------------------------------------------------------//
/**
 * p``悷
 * @param g `悷OtBbNXReLXg
 * @param x <i>x</i> W
 * @param y <i>y</i> W
 * @param value _̒li[z
 * @param s p`W
 */
  protected void fillPolygon(Graphics2D g,int x,int y,
      	      	      	     double value[],PolygonCoordinates s){
    if(s==null) return;
    g.setColor(colorbar.color(s.valueAt(value)));
    g.drawLine(x,y,x,y);
  }

//--------------------------------------------------------------------//
/**
 * ʏ̓_Ԃ
 * @param index ߓ_̃CfbNX
 * @param vCoef f̕\{
 * @param cx fS <i>x</i> W
 * @param cy fS <i>y</i> W
 * @return ʏ̓_
 */
  protected ViewPoint2D getViewPoint(int index,double vCoef,
      	      	      	      	     double cx,double cy){
    try{
      Point2D p=model.getMesh().getNode(index);
      ViewPoint2D vp=new ViewPoint2D(vCoef*(p.getX()-cx),
      	      	      	      	     vCoef*(p.getY()-cy));
      transform.trans(vp);
      return vp;
    }
    catch(Exception ex){
      return null;
    }
  }
}
