package dche.io;
import java.io.BufferedReader;
import dche.Dche;
import dche.Fluid;
import dche.FluidElement;
import dche.DcheDataModel;
import fem.Material;
import fem.MeshModel;
import fem.Node;
import fem.bc.HeatFluxBound;
import fem.bc.HeatTransferBound;
import fem.bc.Temperature;
import fem.element.QuadElement2D;
import fem.element.TriElement2D;
import fileio.TextDataReader;

//--------------------------------------------------------------------//
/**
 * DCHE f[^ǂݍ݃NX
 */
public class DcheModelReader extends TextDataReader{

/** DCHE f[^f */
  protected DcheDataModel model;

/** bVf */
  protected MeshModel mesh;

/**
 * DCHE f[^ǂݍ݃IuWFNg\z
 * @param model DCHE f[^f
 */
  public DcheModelReader(DcheDataModel model){
    this.model=model;
    mesh=model.getMesh();
  }

//--------------------------------------------------------------------//
/**
 * FEM f[^ǂݍ
 * @param br ̓[_[
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  public void readData(BufferedReader br) throws Exception{
    int vertex[]=new int[4];
    model.clear();
    while(true){
      String line=br.readLine();
      if(line==null) break;
      String ss[]=split(line);
      if(ss.length>0){
      	String keyWord=ss[0];
      	if(keyWord.equals("Material")){
      	  int label=Integer.parseInt(ss[1]);
      	  double dens=Double.parseDouble(ss[2]);
      	  double sh=Double.parseDouble(ss[3]);
      	  double hcon=Double.parseDouble(ss[4]);
      	  Material material=new Material(label,dens,sh,hcon);
      	  model.addMaterial(material);
      	}
      	else if(keyWord.equals("Fluid")){
      	  model.addFluid(readFluid(ss,br));
      	}
      	else if(keyWord.equals("Node")){
      	  int label=Integer.parseInt(ss[1]);
      	  double x=Double.parseDouble(ss[2]);
      	  double y=Double.parseDouble(ss[3]);
      	  mesh.addNode(new Node(label,x,y));
      	}
      	else if(keyWord.equals(TriElement2D.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<3;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new TriElement2D(label,material,vertex));
      	}
      	else if(keyWord.equals(QuadElement2D.ELEMENT_NAME)){
      	  int label=Integer.parseInt(ss[1]);
      	  int material=Integer.parseInt(ss[2]);
      	  for(int i=0;i<4;i++){
      	    vertex[i]=Integer.parseInt(ss[3+i]);
      	  }
      	  mesh.addElement(new QuadElement2D(label,material,vertex));
      	}
      	else if(keyWord.equals("Temperature")){
      	  int label=Integer.parseInt(ss[1]);
      	  double t=Double.parseDouble(ss[2]);
      	  model.getBoundaryCondition().addTemperature
      	    (new Temperature(label,t));
      	}
      	else if(keyWord.equals("HTC")){
      	  int label=Integer.parseInt(ss[1]);
      	  double htc=Double.parseDouble(ss[3]);
      	  double temp=Double.parseDouble(ss[4]);
      	  model.getBoundaryCondition().addHTBound
      	    (new HeatTransferBound(label,ss[2],htc,temp));
      	}
      	else if(keyWord.equals("Flux")){
      	  int label=Integer.parseInt(ss[1]);
      	  double flux=Double.parseDouble(ss[3]);
      	  model.getBoundaryCondition().addFluxBound
      	    (new HeatFluxBound(label,ss[2],flux));
      	}
      	else if(keyWord.equals("DCHE")){
      	  model.setDche(readDche(ss,br));
      	}
      	else if(keyWord.equals("Monitor")){
      	  int mon[]=new int[ss.length-1];
      	  for(int i=0;i<ss.length-1;i++){
      	    mon[i]=Integer.parseInt(ss[i+1]);
      	  }
      	  model.setMonitor(mon);
      	}
      }
    }
    model.init();
  }

//--------------------------------------------------------------------//
/**
 * ̕f[^ǂݍ
 * @param ss f[^
 * @param br ̓[_[
 * @return ̕f[^
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  public Fluid readFluid(String ss[],BufferedReader br)
    throws Exception{
    int label=Integer.parseInt(ss[1]);
    int count=Integer.parseInt(ss[2]);
    double t[]=new double[count];
    FluidElement elem[]=new FluidElement[count];
    for(int i=0;i<count;i++){
      String line=br.readLine();
      String s[]=split(line);
      t[i]=Double.parseDouble(s[0]);
      double dens=Double.parseDouble(s[1]);
      double cv=Double.parseDouble(s[2]);
      double lam=Double.parseDouble(s[3]);
      double mu=Double.parseDouble(s[4]);
      double pr=Double.parseDouble(s[5]);
      elem[i]=new FluidElement(dens,cv,lam,mu,pr);
    }
    return new Fluid(label,count,t,elem);
  }

//--------------------------------------------------------------------//
/**
 * DCHE f[^ǂݍ
 * @param ss f[^
 * @param br ̓[_[
 * @return DCHE f[^
 * @throws Exception 炩̗RŃt@Cǂ߂ȂƂ
 */
  public Dche readDche(String ss[],BufferedReader br)
    throws Exception{
    int fl=Integer.parseInt(ss[1]);
    double d1=Double.parseDouble(ss[2]);
    double d2=Double.parseDouble(ss[3]);
    double t=Double.parseDouble(ss[4]);
    double qm=Double.parseDouble(ss[5]);
    double lm=Double.parseDouble(ss[6]);
    String line=br.readLine();
    String s[]=split(line);
    int count=s.length;
    int v1[]=new int[count];
    int v2[]=new int[count];
    int v3[]=new int[count];
    for(int i=0;i<count;i++){
      v1[i]=Integer.parseInt(s[i]);
    }
    line=br.readLine();
    s=split(line);
    for(int i=0;i<count;i++){
      v2[i]=Integer.parseInt(s[i]);
    }
    line=br.readLine();
    s=split(line);
    for(int i=0;i<count;i++){
      v3[i]=Integer.parseInt(s[i]);
    }
    return new Dche(fl,d1,d2,t,qm,lm,count,v1,v2,v3);
  }
}
