package fem;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import fem.bc.BoundaryCondition;
import fem.bc.HeatTransferBound;
import fem.bc.Temperature;
import fem.result.Result;
import figure.Matrix;
import figure.Vector;
import figure.matrix.MatrixSolveException;
import figure.matrix.MatrixSolveMethod;
import figure.matrix.MatrixSolver;
import figure.matrix.SparseMatrixSolver;

//--------------------------------------------------------------------//
/**
 * FEM f[^f
 */
public class FEMDataModel{

/** bVf */
  protected MeshModel mesh;

/** ޗi[郊Xg */
  protected ArrayList<Material> materials;

/** E */
  protected BoundaryCondition bc;

/** vZ */
  protected Result result;

/** f̗L\ boolean ^ */
  protected boolean modelExist;

/** gU}gbNX */
  protected MatrixSolver matrix;

/** xxNg */
  protected Vector vector;

/** s@ */
  protected MatrixSolveMethod method;

/**
 * FEM f[^f\z
 */
  public FEMDataModel(){
    materials=new ArrayList<Material>();
    mesh=new MeshModel();
    bc=new BoundaryCondition();
    result=new Result();
    modelExist=false;
    method=MatrixSolveMethod.ILUCG;
  }

//--------------------------------------------------------------------//
/**
 * bVfԂ
 * @return bVf
 */
  public MeshModel getMesh(){
    return mesh;
  }

//--------------------------------------------------------------------//
/**
 * EԂ
 * @return E
 */
  public BoundaryCondition getBoundaryCondition(){
    return bc;
  }

//--------------------------------------------------------------------//
/**
 * vZʂԂ
 * @return vZ
 */
  public Result getResult(){
    return result;
  }

//--------------------------------------------------------------------//
/**
 * ޗ̐Ԃ
 * @return ޗ̐
 */
  public int materialCount(){
    return materials.size();
  }

//--------------------------------------------------------------------//
/**
 * ޗԂ
 * @param index ޗ̃CfbNX
 * @return w肳ꂽʒu̍ޗ
 */
  public Material getMaterial(int index){
    return materials.get(index);
  }

//--------------------------------------------------------------------//
/**
 * ޗǉ
 * @param material ޗ
 */
  public void addMaterial(Material material){
    materials.add(material);
  }

//--------------------------------------------------------------------//
/**
 * s@ݒ肷
 * @param method s@
 */
  public void setMethod(MatrixSolveMethod method){
    this.method=method;
  }

//--------------------------------------------------------------------//
/**
 * f[^
 */
  public void clear(){
    materials.clear();
    modelExist=false;
    mesh.clear();
    bc.clear();
    clearResults();
  }

//--------------------------------------------------------------------//
/**
 * vZʂ
 */
  public void clearResults(){
    result.clear();
  }

//--------------------------------------------------------------------//
/**
 * f̗L\ boolean ^Ԃ
 * @return f݂ꍇ true
 */
  public boolean modelExist(){
    return modelExist;
  }

//--------------------------------------------------------------------//
/**
 * f
 */
  public void init() throws LabelException{
    materials.trimToSize();
    Collections.sort(materials,MeshModel.LABEL_COMPARATOR);
    mesh.init();
    bc.init();

    reNumbering();
    resetMaterialLabel();
    mesh.getFreeBorder();
    modelExist=true;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_|C^ݒ肷
 * @throws LabelException f[^xvȂƂ
 */
  public void reNumbering() throws LabelException{
    LabelMap labelMap=new LabelMap();
    int nodeCount=mesh.getNodeCount();
    for(int i=0;i<nodeCount;i++){
      labelMap.put(mesh.getNode(i),i);
    }

    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      labelMap.reset((Nodes)mesh.getElement(i));
    }

    int tempCount=bc.temperatureCount();
    for(int i=0;i<tempCount;i++){
      labelMap.reset(bc.getTemperature(i));
    }

    labelMap.clear();
    for(int i=0;i<elementCount;i++){
      labelMap.put(mesh.getElement(i),i);
    }

    int htcCount=bc.htcCount();
    for(int i=0;i<htcCount;i++){
      labelMap.reset(bc.getHTBound(i));
    }
  }

//--------------------------------------------------------------------//
/**
 * ޗ|C^ݒ肷
 * @throws LabelException f[^xvȂƂ
 */
  public void resetMaterialLabel() throws LabelException{
    int materialCount=materials.size();
    if(materialCount==0){
      materials.add(new Material(1));
      return;
    }

    HashMap<Integer,Integer> materialMap=new HashMap<Integer,Integer>();
    for(int i=0;i<materialCount;i++){
      materialMap.put(getMaterial(i).label(),i);
    }

    int elementCount=mesh.getElementCount();
    for(int i=0;i<elementCount;i++){
      Element2D element=mesh.getElement(i);
      Integer newMaterial=materialMap.get(element.getMaterial());
      if(newMaterial!=null){
      	element.setMaterial(newMaterial.intValue());
      }
      else{
      	throw new LabelException("Bad material label");
      }
    }
  }

//--------------------------------------------------------------------//
/**
 * M`BvZ
 */
  public void calculate(){
    createHeatMatrix();
    try{
      Vector temp=matrix.solve(vector,method);
      result.clear();
      result.setTemperature(bc,temp,bc.getTemperature());
    }
    catch(MatrixSolveException ex){
      ex.printStackTrace();
    }
  }

//--------------------------------------------------------------------//
/**
 * MvZ̃}gbNXExNgvZ
 */
  protected void createHeatMatrix(){
    int nodeCount=mesh.getNodeCount();
    int elementCount=mesh.getElementCount();

    int dof=bc.bcList(mesh.getNodeCount());
    int length=nodeCount-dof;
    int ii=0;
    int reducedList[]=new int[length];
    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]<0){
      	reducedList[ii]=i;
      	ii++;
      }
    }

    SparseMatrixSolver matrix1=new SparseMatrixSolver(nodeCount);
    Vector vector1=bc.loadVector(nodeCount);
    for(int i=0;i<elementCount;i++){
      Element2D element=(Element2D)mesh.getElement(i);
      int ncount=element.nodeCount();
      int nodesOnElement[]=element.nodes();
      Material material=materials.get(element.getMaterial());
      Matrix ls=element.gradMatrix(mesh.nodes(element),material);
      for(int i1=0;i1<ncount;i1++){
      	for(int i2=0;i2<ncount;i2++){
      	  matrix1.add(nodesOnElement[i1],nodesOnElement[i2],
      	      	      ls.get(i1,i2));
      	}
      }
    }

    int boundCount=bc.htcCount();
    for(int i=0;i<boundCount;i++){
      HeatTransferBound htc=bc.getHTBound(i);
      ElementBorder border=htc.getBorder(mesh.getElement(htc.label()));
      int ncount=border.nodeCount();
      int nodesOnElement[]=border.nodes();
      Node nodes[]=mesh.nodes(border);
      Matrix hm=border.shapeFunctionMatrix(nodes,htc.htc);
      Vector hv=border.shapeFunctionVector(nodes,htc.outTemp*htc.htc);
      for(int i1=0;i1<ncount;i1++){
      	for(int i2=0;i2<ncount;i2++){
      	  matrix1.add(nodesOnElement[i1],nodesOnElement[i2],
      	      	      hm.get(i1,i2));
      	}
      	vector1.add(nodesOnElement[i1],hv.get(i1));
      }
    }

    for(int i=0;i<bc.bcList.length;i++){
      if(bc.bcList[i]>=0){
      	Temperature temp=bc.getTemperature(bc.bcList[i]);
      	for(int j=0;j<vector1.size;j++){
      	  vector1.add(j,-temp.temperature()*matrix1.get(j,i));
      	}
      }
    }

    matrix=matrix1.extruct(length,reducedList,reducedList);
    vector=new Vector(length);
    for(int i=0;i<length;i++){
      vector.set(i,vector1.get(reducedList[i]));
    }
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNXԂ
 * @return gU}gbNX
 */
  public MatrixSolver getDiffMatrix(){
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * xxNgԂ
 * @return xxNg
 */
  public Vector getTempVector(){
    return vector;
  }
}
