package fem.element;
import java.awt.geom.Point2D;
import java.util.Arrays;
import fem.AbstractNodes;
import fem.Element2D;
import fem.ElementBorder;
import fem.Material;
import fem.Node;
import figure.DefaultMatrix;
import figure.Matrix;

//--------------------------------------------------------------------//
/**
 * Qvf̒ۃX[p[NX
 */
public abstract class AbstractElement2D extends AbstractNodes
  implements Cloneable,Element2D{

/** vfx */
  protected int label;

/** ޗ̃CfbNX */
  protected int material;

/**
 * Qvf\z
 * @param label vfx
 * @param material ޗ̃CfbNX
 */
  protected AbstractElement2D(int label,int material){
    super();
    this.label=label;
    this.material=material;
  }

//--------------------------------------------------------------------//
/**
 * vf̂Ԃ
 * @return vf
 */
  public abstract String getName();

//--------------------------------------------------------------------//
/**
 * vfxԂ
 * @return vfx
 */
  public int label(){
    return label;
  }

//--------------------------------------------------------------------//
/**
 * vfxݒ肷
 * @param label vfx
 */
  public void setLabel(int label){
    this.label=label;
  }

//--------------------------------------------------------------------//
/**
 * ޗ̃CfbNXԂ
 * @return ޗ̃CfbNX
 */
  public int getMaterial(){
    return material;
  }

//--------------------------------------------------------------------//
/**
 * ޗ̃CfbNXݒ肷
 * @param material ޗ̃CfbNX
 */
  public void setMaterial(int material){
    this.material=material;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @param element vfx
 * @return vfEi[z
 */
  public ElementBorder[] borders(int element){
    int borderCount=borderCount();
    ElementBorder borders[]=new ElementBorder[borderCount];
    for(int i=0;i<borderCount;i++) borders[i]=border(element,i);
    return borders;
  }

//--------------------------------------------------------------------//
/**
 * vfߓ_̊pxԂ
 * @param p ߓ_i[z
 * @return vfߓ_̊pxi[z
 */
  public double[] angle(Point2D p[]){
    int nodeCount=nodeCount();
    double angle[]=new double[nodeCount];
    for(int i=0;i<nodeCount;i++){
      Point2D p1=p[(i+1)%nodeCount];
      Point2D p2=p[(i+nodeCount-1)%nodeCount];
      double dx1=p1.getX()-p[i].getX();
      double dy1=p1.getY()-p[i].getY();
      double dx2=p2.getX()-p[i].getX();
      double dy2=p2.getY()-p[i].getY();
      angle[i]=dx1*dx2+dy1*dy2;
      double mag2=(dx1*dx1+dy1*dy1)*(dx2*dx2+dy2*dy2);
      if(mag2>0.0) angle[i]/=Math.sqrt(mag2);
    }
    return angle;
  }

//--------------------------------------------------------------------//
/**
 * RrsԂ
 * @param p ߓ_i[z
 * @param sf `֐s
 * @return Rrs
 */
  protected Matrix jacobianMatrix(Point2D p[],Matrix sf){
    Matrix jacob=new DefaultMatrix(2);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      jacob.add(0,0,dndx2*p[i].getX());
      jacob.add(0,1,dndx2*p[i].getY());
      jacob.add(1,0,dndy2*p[i].getX());
      jacob.add(1,1,dndy2*p[i].getY());
    }
    return jacob;
  }

//--------------------------------------------------------------------//
/**
 * RrAԂ
 * @param ja Rrs
 * @return RrA
 */
  protected double jacobian(Matrix ja){
    return ja.get(0,0)*ja.get(1,1)-ja.get(0,1)*ja.get(1,0);
  }

//--------------------------------------------------------------------//
/**
 * tRrsԂ
 * @param ja Rrs
 * @param jacobian RrA
 * @return tRrs
 */
  protected Matrix jacobInv(Matrix ja,double jacobian){
    Matrix matrix=new DefaultMatrix(2);
    if(jacobian==0.0) return matrix;
    double jinv=1.0/jacobian;
    matrix.set(0,0,jinv*ja.get(1,1));
    matrix.set(0,1,-jinv*ja.get(0,1));
    matrix.set(1,0,-jinv*ja.get(1,0));
    matrix.set(1,1,jinv*ja.get(0,0));
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * `֐̌z<br>
 * { <i>dN<sub>i</sub></i> /<i>dx</i>
 *   <i>dN<sub>i</sub></i> /<i>dy</i>
 *   <i>dN<sub>i</sub></i> /<i>dz</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐̌zi[z
 */
  protected double[][] grad(Point2D p[],double xsi,double eta){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    return grad(p,ja,jacobian,sf);
  }

//--------------------------------------------------------------------//
/**
 * `֐̌z<br>
 * { <i>dN<sub>i</sub></i> /<i>dx</i>
 *   <i>dN<sub>i</sub></i> /<i>dy</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param ja Rrs
 * @param jacobian RrA
 * @param sf `֐s
 * @return `֐̌zi[z
 */
  protected double[][] grad(Point2D p[],Matrix ja,double jacobian,
      	      	      	    Matrix sf){
    int nodeCount=nodeCount();
    double jinv=1.0/jacobian;
    double j11=jinv*ja.get(1,1);
    double j12=-jinv*ja.get(0,1);
    double j21=-jinv*ja.get(1,0);
    double j22=jinv*ja.get(0,0);
    double grad[][]=new double[nodeCount][2];
    for(int i=0;i<nodeCount;i++){
      double dndx2=sf.get(i,1);
      double dndy2=sf.get(i,2);
      grad[i][0]=j11*dndx2+j12*dndy2;
      grad[i][1]=j21*dndx2+j22*dndy2;
    }
    return grad;
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param material ޗ
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point2D p[],Material material){
    return shapeFunctionMatrix(p,material.cv);
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param material ޗ
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point2D p[],Material material){
    return gradMatrix(p,material.heatConductivity);
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub></i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param w d݌W
 * @return `֐}gbNX
 */
  protected Matrix shapeFunctionMatrix(Point2D p[],double xsi,
      	      	      	      	       double eta,double w){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(nodeCount);
    double coef=w*Math.abs(jacobian);
    for(int i=0;i<nodeCount;i++){
      for(int j=0;j<nodeCount;j++){
      	matrix.set(i,j,coef*sf.get(i,0)*sf.get(j,0));
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub></i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @param w d݌W
 * @return gU}gbNX
 */
  protected Matrix gradMatrix(Point2D p[],double xsi,double eta,
      	      	      	      double w){
    Matrix sf=shapeFunction(xsi,eta);
    Matrix ja=jacobianMatrix(p,sf);
    double jacobian=jacobian(ja);
    double grad[][]=grad(p,ja,jacobian,sf);
    int nodeCount=nodeCount();
    Matrix matrix=new DefaultMatrix(nodeCount);
    double coef=w*Math.abs(jacobian);
    for(int i=0;i<nodeCount;i++){
      for(int j=0;j<nodeCount;j++){
      	matrix.set(i,j,coef*(grad[i][0]*grad[j][0]
      	      	      	    +grad[i][1]*grad[j][1]));
      }
    }
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃Rs[Ԃ
 * @return ̃CX^X̕
 */
  public Object clone(){
    return copyElement();
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃Rs[Ԃ
 * @return ̃CX^X̕
 */
  public Element2D copyElement(){
    try{
      AbstractElement2D element=(AbstractElement2D)super.clone();
      element.nodes=new int[nodes.length];
      System.arraycopy(nodes,0,element.nodes,0,nodes.length);
      return element;
    }
    catch(CloneNotSupportedException ex){
      throw new InternalError();
    }
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̃nbVR[hlԂ
 * @return IuWFNg̃nbVR[hl
 */
  public int hashCode(){
    int bit=31*label+material;
    for(int i=0;i<nodes.length;i++){
      bit=31*bit+nodes[i];
    }
    return bit;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @param obj rΏۂ̃IuWFNg
 * @return rΏۂ̃IuWFNg <code>AbstractElement2D</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof AbstractElement2D){
      AbstractElement2D elem=(AbstractElement2D)obj;
      int nodes[]=elem.nodes();
      return ((label==elem.label)
      	   && (material==elem.material)
      	   && Arrays.equals(nodes,elem.nodes));
    }
    else{
      return false;
    }
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @return ̃IuWFNg̕\
 */
  public String toString(){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(label);
    sb.append('\t');
    sb.append(material);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(nodes[i]);
    }
    return sb.toString();
  }

//--------------------------------------------------------------------//
/**
 * f[^Ԃ
 * @param material ޗx
 * @param p ߓ_i[z
 * @return ̃IuWFNg̕\
 */
  public String toString(int material,Node p[]){
    StringBuilder sb=new StringBuilder(getName());
    sb.append('\t');
    sb.append(label);
    sb.append('\t');
    sb.append(material);
    int nodeCount=nodeCount();
    for(int i=0;i<nodeCount;i++){
      sb.append('\t');
      sb.append(p[i].label());
    }
    return sb.toString();
  }
}
