package fem.element;
import java.awt.geom.Point2D;
import fem.border.EdgeBorder;
import fem.ElementBorder;
import figure.DefaultMatrix;
import figure.Matrix;

//--------------------------------------------------------------------//
/**
 * lp` 1 vfNX
 */
public class QuadElement2D extends AbstractElement2D{

/** vf */
  public static final String ELEMENT_NAME="QuadElement2D";

/** ߓ_ */
  public static final int NODE_COUNT=4;

/** vfE */
  public static final int BORDER_COUNT=4;

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param p1 ߓ_ԍ
 * @param p2 ߓ_ԍ
 * @param p3 ߓ_ԍ
 * @param p4 ߓ_ԍ
 */
  public QuadElement2D(int label,int material,int p1,int p2,
      	      	       int p3,int p4){
    super(label,material);
    nodes=new int[4];
    nodes[0]=p1;
    nodes[1]=p2;
    nodes[2]=p3;
    nodes[3]=p4;
  }

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param p ߓ_ԍi[z
 */
  public QuadElement2D(int label,int material,int p[]){
    this(label,material,p[0],p[1],p[2],p[3]);
  }

//--------------------------------------------------------------------//
/**
 * vf̂Ԃ
 * @return vf
 */
  public String getName(){
    return ELEMENT_NAME;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_Ԃ
 * @return ߓ_
 */
  public int nodeCount(){
    return NODE_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @return vfE
 */
  public int borderCount(){
    return BORDER_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @param element vfx
 * @param index vfẼCfbNX
 * @return vfEi[z
 */
  public ElementBorder border(int element,int index){
    switch(index){
      default:
      	return null;
      case 0:
      	return new EdgeBorder(element,nodes[0],nodes[1]);
      case 1:
      	return new EdgeBorder(element,nodes[1],nodes[2]);
      case 2:
      	return new EdgeBorder(element,nodes[2],nodes[3]);
      case 3:
      	return new EdgeBorder(element,nodes[3],nodes[0]);
    }
  }

//--------------------------------------------------------------------//
/**
 * `֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐s
 */
  public Matrix shapeFunction(double xsi,double eta){
    Matrix m=new DefaultMatrix(NODE_COUNT,3);
    m.set(0,0,0.25*(1.0-xsi)*(1.0-eta));
    m.set(1,0,0.25*(1.0+xsi)*(1.0-eta));
    m.set(2,0,0.25*(1.0+xsi)*(1.0+eta));
    m.set(3,0,0.25*(1.0-xsi)*(1.0+eta));
    m.set(0,1,-0.25*(1.0-eta));
    m.set(1,1,0.25*(1.0-eta));
    m.set(2,1,0.25*(1.0+eta));
    m.set(3,1,-0.25*(1.0+eta));
    m.set(0,2,-0.25*(1.0-xsi));
    m.set(1,2,-0.25*(1.0+xsi));
    m.set(2,2,0.25*(1.0+xsi));
    m.set(3,2,0.25*(1.0-xsi));
    return m;
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point2D p[],double coef){
    Matrix s=new DefaultMatrix(NODE_COUNT);
    double r=0.0;
    for(int i=0;i<NODE_COUNT;i++){		// KEXϕ
      s.add(shapeFunctionMatrix(p,GX2[i%2],GX2[(i/2)%2],coef));
      r+=p[i].getX();
    }
    s.mul(r/(double)NODE_COUNT);
    return s;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point2D p[],double coef){
    Matrix g=new DefaultMatrix(NODE_COUNT);
    double r=0.0;
    for(int i=0;i<NODE_COUNT;i++){		// KEXϕ
      g.add(gradMatrix(p,GX2[i%2],GX2[(i/2)%2],coef));
      r+=p[i].getX();
    }
    g.mul(r/(double)NODE_COUNT);
    return g;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @return rΏۂ̃IuWFNg <code>QuadElement2D</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof QuadElement2D){
      return super.equals(obj);
    }
    else{
      return false;
    }
  }
}
