package fem.element;
import java.awt.geom.Point2D;
import fem.border.EdgeBorder;
import fem.ElementBorder;
import figure.DefaultMatrix;
import figure.Matrix;

//--------------------------------------------------------------------//
/**
 * Op` 1 vfNX
 */
public class TriElement2D extends AbstractElement2D{

/** vf */
  public static final String ELEMENT_NAME="TriElement2D";

/** ߓ_ */
  public static final int NODE_COUNT=3;

/** vfE */
  public static final int BORDER_COUNT=3;

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param p1 ߓ_ԍ
 * @param p2 ߓ_ԍ
 * @param p3 ߓ_ԍ
 */
  public TriElement2D(int label,int material,int p1,int p2,int p3){
    super(label,material);
    nodes=new int[3];
    nodes[0]=p1;
    nodes[1]=p2;
    nodes[2]=p3;
  }

/**
 * vf\z
 * @param label vfx
 * @param material ޗԍ
 * @param p ߓ_ԍi[z
 */
  public TriElement2D(int label,int material,int p[]){
    this(label,material,p[0],p[1],p[2]);
  }

//--------------------------------------------------------------------//
/**
 * vf̂Ԃ
 * @return vf
 */
  public String getName(){
    return ELEMENT_NAME;
  }

//--------------------------------------------------------------------//
/**
 * ߓ_Ԃ
 * @return ߓ_
 */
  public int nodeCount(){
    return NODE_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @return vfE
 */
  public int borderCount(){
    return BORDER_COUNT;
  }

//--------------------------------------------------------------------//
/**
 * vfEԂ
 * @param element vfx
 * @param index vfẼCfbNX
 * @return vfEi[z
 */
  public ElementBorder border(int element,int index){
    switch(index){
      default:
      	return null;
      case 0:
      	return new EdgeBorder(element,nodes[0],nodes[1]);
      case 1:
      	return new EdgeBorder(element,nodes[1],nodes[2]);
      case 2:
      	return new EdgeBorder(element,nodes[2],nodes[0]);
    }
  }

//--------------------------------------------------------------------//
/**
 * `֐s<br>
 * { <i>N<sub>i</sub></i>
 *   <i>dN<sub>i</sub></i> /<i>d</i>
 *   <i>dN<sub>i</sub></i> /<i>d</i> }<br>
 * Ԃ
 * @param xsi <i></i> W
 * @param eta <i></i> W
 * @return `֐s
 */
  public Matrix shapeFunction(double xsi,double eta){
    Matrix m=new DefaultMatrix(NODE_COUNT,3);
    m.set(0,0,1.0-xsi-eta);
    m.set(1,0,xsi);
    m.set(2,0,eta);
    m.set(0,1,-1.0);
    m.set(1,1,1.0);
    m.set(0,2,-1.0);
    m.set(2,2,1.0);
    return m;
  }

//--------------------------------------------------------------------//
/**
 * RrAԂ
 * @param p ߓ_i[z
 * @return RrA
 */
  public double jacobian(Point2D p[]){
    return Math.abs((p[1].getX()-p[0].getX())
      	      	   *(p[2].getY()-p[0].getY())
      	      	   -(p[1].getY()-p[0].getY())
      	      	   *(p[2].getX()-p[0].getX()));
  }

//--------------------------------------------------------------------//
/**
 * `֐}gbNX<br>
 * { <i>N<sub>i</sub>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return `֐}gbNX
 */
  public Matrix shapeFunctionMatrix(Point2D p[],double coef){
    Matrix s=new DefaultMatrix(NODE_COUNT);
    double value=coef*jacobian(p)/12.0;
    double r=0.0;
    for(int i=0;i<NODE_COUNT;i++){		// KEXϕ
      s.set(i,i,value);
      for(int j=i+1;j<NODE_COUNT;j++){
      	s.set(i,j,0.5*value);
      	s.set(j,i,0.5*value);
      }
      r+=p[i].getX();
    }
    s.mul(r/(double)NODE_COUNT);
    return s;
  }

//--------------------------------------------------------------------//
/**
 * gU}gbNX<br>
 * { <i>N<sub>i</sub></i>E<i>N<sub>j</sub> dV</i> }<br>
 * Ԃ
 * @param p ߓ_i[z
 * @param coef W
 * @return gU}gbNX
 */
  public Matrix gradMatrix(Point2D p[],double coef){
    Matrix g=new DefaultMatrix(NODE_COUNT);
    double jacob=Math.abs(jacobian(p));
    double ji=1.0/jacob;
    double dx[]={ji*(p[1].getY()-p[2].getY()),
      	      	 ji*(p[2].getY()-p[0].getY()),
      	      	 ji*(p[0].getY()-p[1].getY())};
    double dy[]={ji*(p[2].getX()-p[1].getX()),
      	      	 ji*(p[0].getX()-p[2].getX()),
      	      	 ji*(p[1].getX()-p[0].getX())};
    double value=coef*jacob;
    double r=0.0;
    for(int i=0;i<NODE_COUNT;i++){		// KEXϕ
      for(int j=0;j<NODE_COUNT;j++){
      	g.set(i,j,value*(dx[i]*dx[j]+dy[i]*dy[j]));
      }
      r+=p[i].getX();
    }
    g.mul(r/(double)NODE_COUNT);
    return g;
  }

//--------------------------------------------------------------------//
/**
 * IuWFNg̓
 * @return rΏۂ̃IuWFNg <code>TriElement2D</code>
 *         ̃CX^Xœlꍇ true
 */
  public boolean equals(Object obj){
    if(obj instanceof TriElement2D){
      return super.equals(obj);
    }
    else{
      return false;
    }
  }
}
