package figure.matrix;
import java.util.ArrayList;
import java.util.Arrays;
import data.IntegerList;
import figure.Matrix;
import figure.SparseMatrix;
import figure.Vector;

//--------------------------------------------------------------------//
/**
 * Ả߂邽߂̑asNX
 */
public class SparseMatrixSolver
  extends SparseMatrix implements MatrixSolver{

/**
 * s\z
 * @param size s̑傫
 */
  public SparseMatrixSolver(int size){
    super(size);
  }

//--------------------------------------------------------------------//
/**
 * w肵sEosԂ
 * @param size s̑傫
 * @param rows s̃CfbNXi[z
 * @param columns ̃CfbNXi[z
 * @return ̍s񂩂w肳ꂽsEos
 */
  public Matrix extructRowColumn(int size,int rows[],int columns[]){
    return extruct(size,rows,columns);
  }

//--------------------------------------------------------------------//
/**
 * w肵sEosԂ
 * @param size s̑傫
 * @param rows s̃CfbNXi[z
 * @param columns ̃CfbNXi[z
 * @return ̍s񂩂w肳ꂽsEos
 */
  public MatrixSolver extruct(int size,int rows[],int columns[]){
    MatrixSolver matrix=new SparseMatrixSolver(size);
    for(int i=0;i<size;i++){
      int columnCount=rowdata[rows[i]].size();
      int j1=0,j2=0;
      while(j1<columnCount && j2<size){
      	Data data=getData(rows[i],j1);
      	while((j2<size) && (columns[j2]<data.column)) j2++;
      	if((j2<size) && (columns[j2]==data.column)){
      	  matrix.set(i,j2,data.value);
      	}
      	j1++;
      }
    }
    matrix.init();
    return matrix;
  }

//--------------------------------------------------------------------//
/**
 * Ả߂
 * @param p xNg <i>p</i>
 * @param method s@
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solve(Vector p,MatrixSolveMethod method)
    throws MatrixSolveException{

    double param[]=MatrixSolveParameter.getParameters(method);
    switch(method){
      case GAUSS:
      	return solveByGauss(p,param[0]);
      case GAUSS_JORDAN:
      	return solveByGaussJordan(p,param[0]);
      case LU:
      	return solveByLU(p,param[0]);
      case SOR:
      	return solveBySOR(p,param[0],(int)param[1],param[2]);
      case CG:
      	return solveByCG(p,param[0],(int)param[1]);
      case ILUCG:
      	return solveByILUCG(p,param[0],(int)param[1]);
    }
    return null;
  }

//--------------------------------------------------------------------//
/**
 * KEX@ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByGauss(Vector p,double tolerance)
    throws MatrixSolveException{

    int size=p.size();
    tolerance*=dataMax;
    SparseMatrix m=(SparseMatrix)clone();
    Vector x=(Vector)p.clone();
    for(int i=0;i<size-1;i++){
      if(Math.abs(m.get(i,i))<tolerance) return null;
      double aii=1.0/m.get(i,i);
      for(int j=i+1;j<rowdata.length;j++){
      	double aji=-aii*m.get(j,i);
      	int columnCount=m.columnCount(i);
      	for(int k=0;k<columnCount;k++){
      	  Data data=m.getData(i,k);
      	  if(data.column>i) m.add(j,data.column,aji*data.value);
      	}
      	x.add(j,aji*x.get(i));
      }
      int columnCount=m.columnCount(i);
      for(int j=0;j<columnCount;j++){
      	Data data=m.getData(i,j);
      	if(data.column>i) m.set(i,data.column,aii*data.value);
      }
      x.set(i,aii*x.get(i));
    }
    if(Math.abs(m.get(size-1,size-1))<tolerance){
      throw new MatrixSolveException();
    }
    x.set(size-1,x.get(size-1)/m.get(size-1,size-1));
    for(int i=size-2;i>=0;i--){
      double s=0.0;
      for(int j=i+1;j<size;j++){
      	s-=m.get(i,j)*x.get(j);
      }
      x.add(i,s);
    }
    return x;
  }

//--------------------------------------------------------------------//
/**
 * Gauss - Jordan @ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByGaussJordan(Vector p,double tolerance)
    throws MatrixSolveException{

    init();
    int size=p.size();
    tolerance*=dataMax;
    SparseMatrix m=(SparseMatrix)clone();
    Vector x=(Vector)p.clone();
    int index[]=new int[size];
    for(int i=0;i<size;i++) index[i]=i;

    for(int i=0;i<size;i++){
      double amax=0.0;
      int kmax=i;
      int columnCount=m.columnCount(i);
      for(int k=0;k<columnCount;k++){
      	Data data=m.getData(i,k);
      	if((data.column>=i) && (Math.abs(data.value)>amax)){
      	  amax=Math.abs(data.value);
      	  kmax=data.column;
      	}
      }
      if(amax<tolerance){
      	throw new MatrixSolveException();
      }
      if(i!=kmax){
      	for(int j=0;j<size;j++){
      	  amax=m.get(j,i);
      	  m.set(j,i,m.get(j,kmax));
      	  m.set(j,kmax,amax);
      	}
      	int itemp=index[i];
      	index[i]=index[kmax];
      	index[kmax]=itemp;
      }
      double minv=1.0/m.get(i,i);
      for(int k=0;k<columnCount;k++){
      	Data data=m.getData(i,k);
      	if(data.column>i) data.mul(minv);
      }
      x.set(i,minv*x.get(i));
      for(int j=0;j<size;j++){
      	if(i!=j){
      	  double mik=-m.get(j,i);
      	  if(mik!=0.0){
      	    for(int k=0;k<columnCount;k++){
      	      Data data=m.getData(i,k);
      	      if(data.column>i) m.add(j,data.column,mik*data.value);
      	    }
      	    x.add(j,mik*x.get(i));
      	  }
      	}
      }
    }
    Vector x2=new Vector(size);
    for(int i=0;i<size;i++) x2.set(index[i],x.get(i));
    return x2;
  }

//--------------------------------------------------------------------//
/**
 * LU 
 * @param tolerance 臒l
 * @return ̍s LU s
 * @throws MatrixSolveException LU łȂƂ
 */
  public Matrix getLU(double tolerance) throws MatrixSolveException{
    SparseMatrixSolver lu=(SparseMatrixSolver)clone();
    tolerance*=dataMax;
    for(int k=0;k<rowdata.length-1;k++){
      int luColumnCount=lu.columnCount(k);
      int dcol=lu.getDiagonalColumn(k);
      double diag=lu.getDiagonal(k);
      if(Math.abs(diag)<tolerance){
      	throw new MatrixSolveException();
      }
      double dkk=1.0/diag;
      for(int j=luColumnCount-1;j>=dcol+1;j--){
      	Data data=lu.getData(k,j);
      	lu.set(k,data.column,dkk*data.value);
      }
      luColumnCount=lu.columnCount(k);
      for(int i=k+1;i<rowdata.length;i++){
      	double aik=-lu.get(i,k);
      	if(aik!=0.0){
      	  for(int j=dcol+1;j<luColumnCount;j++){
      	    Data data=lu.getData(k,j);
      	    lu.add(i,data.column,aik*data.value);
      	  }
      	}
      }
      lu.init();
    }
    return lu;
  }

//--------------------------------------------------------------------//
/**
 * LU @ŘẢ߂
 * @param lu LU ꂽs
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public static Vector solveByLU(SparseMatrix lu,Vector p,
      	      	      	      	 double tolerance)
    throws MatrixSolveException{

    int size=p.size();
    tolerance*=lu.dataMax;
    double q[]=new double[size];
    if(Math.abs(lu.getDiagonal(0))<tolerance){
      throw new MatrixSolveException();
    }
    q[0]=p.get(0)/lu.getDiagonal(0);
    for(int i=1;i<size;i++){
      double p2=p.get(i);
      int columnCount=lu.columnCount(i);
      for(int j=0;j<columnCount-1;j++){
      	Data data=lu.getData(i,j);
      	if(data.column<i) p2-=data.value*q[data.column];
      }
      if(Math.abs(lu.getDiagonal(i))<tolerance){
      	throw new MatrixSolveException();
      }
      q[i]=p2/lu.getDiagonal(i);
    }
    for(int i=size-2;i>=0;i--){
      int columnCount=lu.columnCount(i);
      for(int j=1;j<columnCount;j++){
      	Data data=lu.getData(i,j);
      	if(data.column>i) q[i]-=data.value*q[data.column];
      }
    }
    return new Vector(size,q);
  }

//--------------------------------------------------------------------//
/**
 * LU @ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByLU(Vector p,double tolerance)
    throws MatrixSolveException{

    SparseMatrixSolver m=(SparseMatrixSolver)clone();
    return m.solveByLUNoPreserv(p,tolerance);
  }

//--------------------------------------------------------------------//
/**
 * LU @ŘẢ߂<br>
 * ̍s͕ێȂ
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByLUNoPreserv(Vector p,double tolerance)
    throws MatrixSolveException{

    init();
    int size=p.size();
    tolerance*=dataMax;

    for(int k=0;k<rowdata.length-1;k++){
      int luColumnCount=columnCount(k);
      int dcol=getDiagonalColumn(k);
      double diag=getDiagonal(k);
      if(Math.abs(diag)<tolerance){
      	throw new MatrixSolveException();
      }
      double akk=1.0/diag;
      for(int j=luColumnCount-1;j>=dcol+1;j--){
      	Data data=getData(k,j);
      	set(k,data.column,akk*data.value);
      }
      luColumnCount=columnCount(k);
      for(int i=k+1;i<rowdata.length;i++){
      	double aik=-get(i,k);
      	if(aik!=0.0){
      	  for(int j=dcol+1;j<luColumnCount;j++){
      	    Data data=getData(k,j);
      	    add(i,data.column,aik*data.value);
      	  }
      	}
      }
      init();
    }

    double q[]=new double[size];
    if(Math.abs(getDiagonal(0))<tolerance){
      throw new MatrixSolveException();
    }
    q[0]=p.get(0)/getDiagonal(0);
    for(int i=1;i<size;i++){
      double p2=p.get(i);
      int columnCount=columnCount(i);
      for(int j=0;j<columnCount-1;j++){
      	Data data=getData(i,j);
      	if(data.column<i) p2-=data.value*q[data.column];
      }
      double diag=getDiagonal(i);
      if(Math.abs(diag)<tolerance){
      	throw new MatrixSolveException();
      }
      q[i]=p2/diag;
    }
    for(int i=size-2;i>=0;i--){
      int columnCount=columnCount(i);
      for(int j=1;j<columnCount;j++){
      	Data data=getData(i,j);
      	if(data.column>i) q[i]-=data.value*q[data.column];
      }
    }
    return new Vector(size,q);
  }

//--------------------------------------------------------------------//
/**
 * SOR @ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @param iterMax 񐔂̏
 * @param relax ɘaW
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveBySOR(Vector p,double tolerance,int iterMax,
      	      	      	   double relax)
    throws MatrixSolveException{

    double tol2=tolerance*dataMax;
    int size=p.size();
    Vector x=new Vector(size);
    SparseMatrix mat=new SparseMatrix(rowdata.length);
    double b[]=new double[size];
    for(int i=0;i<size;i++){
      double dd=get(i,i);
      if(Math.abs(dd)<tol2){
      	throw new MatrixSolveException();
      }
      dd=1.0/dd;
      int columnCount=rowdata[i].size();
      for(int j=0;j<columnCount;j++){
      	Data data=getData(i,j);
      	mat.set(i,data.column,dd*data.value);
      }
      b[i]=dd*p.get(i);
    }
    for(int iter=0;iter<iterMax;iter++){
      double eps1=0.0;
      double eps2=0.0;
      for(int i=0;i<size;i++){
      	double xnew=b[i];
      	int columnCount=mat.columnCount(i);
      	for(int j=0;j<columnCount;j++){
      	  Data data=mat.getData(i,j);
      	  if(i!=data.column) xnew-=data.value*x.get(data.column);
      	}
      	double xold=x.get(i);
      	xnew=xold+relax*(xnew-xold);
      	x.set(i,xnew);
      	eps1+=(xnew-xold)*(xnew-xold);
      	eps2+=xnew*xnew;
      }
      if(eps1<tolerance*eps2) return x;
    }
    return x;
  }

//--------------------------------------------------------------------//
/**
 * z@ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByCG(Vector p,double tolerance)
    throws MatrixSolveException{
    return solveByCG(p,tolerance,p.size());
  }

//--------------------------------------------------------------------//
/**
 * z@ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @param iterMax 񐔂̏
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByCG(Vector p,double tolerance,int iterMax)
    throws MatrixSolveException{

    double tol1=tolerance*dataMax;
    double tol2=tolerance*Vector.innerProduct(p,p);
    int size=p.size();
    Vector x=new Vector(size);
    Vector xg=(Vector)p.clone();
    Vector xd=(Vector)xg.clone();

    iterMax=Math.min(iterMax,rowdata.length);
    for(int iter=0;iter<iterMax;iter++){
      Vector xq=mul(xd);
      double r=Vector.innerProduct(xd,xq);
      if(Math.abs(r)<tol1){
      	throw new MatrixSolveException();
      }
      double alpha=Vector.innerProduct(xd,xg)/r;
      x.add(alpha,xd);
      xg.add(-alpha,xq);
      if(Vector.innerProduct(xg,xg)<tol2) return x;
      double beta=-Vector.innerProduct(xg,xq)/r;
      xd.mul(beta);
      xd.add(xg);
    }
    return x;
  }

//--------------------------------------------------------------------//
/**
 * sS LU z@ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByILUCG(Vector p,double tolerance)
    throws MatrixSolveException{
    return solveByILUCG((SparseMatrix)getILU(tolerance),p,tolerance,
      	      	      	p.size());
  }

//--------------------------------------------------------------------//
/**
 * sS LU z@ŘẢ߂
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @param iterMax 񐔂̏
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByILUCG(Vector p,double tolerance,int iterMax)
    throws MatrixSolveException{
    return solveByILUCG((SparseMatrix)getILU(tolerance),p,tolerance,
      	      	      	iterMax);
  }

//--------------------------------------------------------------------//
/**
 * sS LU z@ŘẢ߂
 * @param lu sS LU ꂽs
 * @param p xNg <i>p</i>
 * @param tolerance 臒l
 * @param iterMax 񐔂̏
 * @return Ả\xNg
 * @throws MatrixSolveException Ả܂ȂƂ
 */
  public Vector solveByILUCG(SparseMatrix lu,Vector p,double tolerance,
      	      	      	     int iterMax)
    throws MatrixSolveException{

    double tol1=tolerance*Vector.innerProduct(p,p);
    double tol2=tolerance*tol1;
    int size=p.size();
    Vector x=new Vector(size);
    Vector xg=(Vector)p.clone();
    Vector xq=solveByLU(lu,xg,tolerance);
    Vector xd=(Vector)xq.clone();
    iterMax=Math.min(iterMax,lu.rowCount());
    for(int iter=0;iter<iterMax;iter++){
      double z1=Vector.innerProduct(xd,xg);
      xq=mul(xd);
      double r=Vector.innerProduct(xd,xq);
      if(Math.abs(r)<tol2){
      	throw new MatrixSolveException();
      }
      double alpha=z1/r;
      x.add(alpha,xd);
      xg.add(-alpha,xq);
      if(Vector.innerProduct(xg,xg)<tol1) return x;
      Vector xq2=solveByLU(lu,xg,tolerance);
      double z2=Vector.innerProduct(xq,xq2);
      double beta=-z2/r;
      xd.mul(beta);
      xd.add(xq2);
    }
    return x;
  }

//--------------------------------------------------------------------//
/**
 * sS LU 
 * @param tolerance 臒l
 * @return ̍ssS LU s
 * @throws MatrixSolveException sS LU łȂƂ
 */
  public Matrix getILU(double tolerance) throws MatrixSolveException{
    init();
    tolerance*=dataMax;

// 񂩂̌pf[^쐬
    IntegerList columnData[]=new IntegerList[rowdata.length];
    int diacol[]=new int[rowdata.length];
    for(int i=0;i<rowdata.length;i++){
      columnData[i]=new IntegerList();
      diacol[i]=-1;
    }
    for(int i=0;i<rowdata.length;i++){
      int columnCount=columnCount(i);
      int dcol=getDiagonalColumn(i);
      for(int j=0;j<columnCount;j++){
      	Data data=getData(i,j);
      	if(j==dcol) diacol[data.column]=columnData[data.column].size();
      	columnData[data.column].add(i);
      }
    }

// sSLU
    SparseMatrixSolver lu=(SparseMatrixSolver)clone();
    for(int k=0;k<rowdata.length-1;k++){
      int luColumnCount=lu.columnCount(k);
      double diag=lu.getDiagonal(k);
      if(Math.abs(diag)<tolerance){
      	throw new MatrixSolveException();
      }
      double dkk=1.0/diag;
      int dcol=lu.getDiagonalColumn(k);
      RowData lulist1=lu.getRowData(k);
      for(int j=dcol+1;j<luColumnCount;j++){
      	Data data=lulist1.getData(j);
      	data.mul(dkk);
      }
      int rowCount=columnData[k].size();
      for(int i=diacol[k]+1;i<rowCount;i++){
      	int row=columnData[k].get(i);
      	double aik=-lu.get(row,k);
      	RowData lulist2=lu.getRowData(row);
      	for(int j=dcol+1;j<luColumnCount;j++){
      	  Data data=lulist1.getData(j);
      	  int col=lulist2.searchColumn(data.column);
      	  if(col>=0){
      	    Data data1=lulist2.getData(col);
      	    if(data1.column==data.column){
      	      data1.add(aik*data.value);
      	    }
      	  }
      	}
      }
    }
    lu.init();
    return lu;
  }

//--------------------------------------------------------------------//
/**
 * ts߂
 * @return ̍s̋ts
 * @throws MatrixSolveException ts񂪋܂ȂƂ
 */
  public MatrixSolver inverse() throws MatrixSolveException{
    return inverse((SparseMatrix)getLU(TOLERANCE),TOLERANCE);
  }

//--------------------------------------------------------------------//
/**
 * LU @ŋts߂
 * @param lu LU ꂽs
 * @param tolerance 臒l
 * @return ̍s̋ts
 * @throws MatrixSolveException ts񂪋܂ȂƂ
 */
  public static MatrixSolver inverse(SparseMatrix lu,double tolerance)
    throws MatrixSolveException{

    tolerance*=lu.dataMax;
    int size=lu.rowCount();
    SparseMatrixSolver ai=new SparseMatrixSolver(size);
    for(int i=0;i<size;i++){
      if(Math.abs(lu.get(i,i))<tolerance){
      	throw new MatrixSolveException();
      }
    }
    ai.set(size-1,size-1,1.0/lu.get(size-1,size-1));
    for(int i=size-2;i>=0;i--){
      for(int j=i;j>=0;j--){
      	double a1=0.0;
      	double a2=0.0;
      	for(int k=j+1;k<size;k++){
      	  a1-=lu.get(j,k)*ai.get(k,i+1);
      	  a2-=ai.get(i+1,k)*lu.get(k,j);
      	}
      	ai.set(j,i+1,a1);
      	ai.set(i+1,j,a2/lu.get(j,j));
      }
      double a1=1.0;
      for(int k=i+1;k<size;k++){
      	a1-=ai.get(i,k)*lu.get(k,i);
      }
      ai.set(i,i,a1/lu.get(i,i));
    }
    ai.init();
    return ai;
  }
}
